@@ -1,4 +1,3 @@
-from __future__ import unicode_literals
 import logging
 from pythonagent.lib import get_ident
 import ctypes
@@ -29,6 +28,7 @@ class TransactionContext(object):
         self.instance_name = None
         self.url_path = None
         self.flow_path_id = None
+        self.havoc_applied = False
 
 
 class Agent(object):
@@ -91,10 +91,15 @@ class Agent(object):
             self.instance_name = "default"
             self.function_name = "default"
 
+        self.nd_cookie_key = None
+        self.nv_cookie_key = None
+
         self.start()
+
         if self.cav_env == "NATIVE":
             self.ai_wrapper_begin()
             self.ad_wrapper()
+            self.register_nd_cookie_callback()
         # self.thd_wrapper()
         # self.ai_wrapper_end()
         # self.ai_wrapper_begin()
@@ -108,6 +113,7 @@ class Agent(object):
 
         if self.cav_env == "NATIVE":
             self.lib = self.load_common_lib(path)
+            self.register_nv_cookie_callback()
 
         #self.logger.info("going to start sdk_init()")
         self.sdk_init()
@@ -146,7 +152,7 @@ class Agent(object):
         self.lib.nd_bt_begin.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
         self.lib.nd_bt_begin.restype = ctypes.c_void_p
 
-        self.lib.nd_bt_end.argtypes = [ctypes.c_void_p, ctypes.c_int]
+        self.lib.nd_bt_end.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
         self.lib.nd_bt_end.restype = ctypes.c_int
 
         self.lib.nd_bt_store.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
@@ -163,9 +169,15 @@ class Agent(object):
         self.lib.nd_ip_http_callout_end.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
         self.lib.nd_ip_http_callout_end.restype = ctypes.c_int
 
+        self.lib.sdk_getTopoFpStr.argtypes = [ctypes.c_void_p]
+        self.lib.sdk_getTopoFpStr.restype = ctypes.c_char_p
+
         self.lib.handleUploadDownload.argtypes = [ctypes.c_char_p]
         self.lib.handleUploadDownload.restype = ctypes.c_void_p
 
+        self.lib.sdk_getNDSessionCookie.argtypes = [ctypes.c_void_p]
+        self.lib.sdk_getNDSessionCookie.restype = ctypes.c_char_p
+
         # self.lib.register_AI_end_callback.argtypes = [ctypes.c_char_p]
 
         CB_FTYPE_DOUBLE_DOUBLE = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_int, ctypes.c_longlong,
@@ -189,18 +201,67 @@ class Agent(object):
 
         #self.lib.sdk_get_otl_tracestate.argtypes = [ctypes.c_void_p]
         #self.lib.sdk_get_otl_tracestate.restype = ctypes.c_char_p
-        
+
+        ftype_nd_cookie = ctypes.CFUNCTYPE(ctypes.c_void_p,    # Pointer to Python Function
+                                           ctypes.c_char_p,    # char *name
+                                           ctypes.c_char_p,    # char *domain
+                                           ctypes.c_int,       # int rheader
+                                           ctypes.c_int,       # int mpos
+                                           ctypes.c_longlong)  # long long int duration
+
+        self.nd_cookie_callback_c = ftype_nd_cookie(self.nd_cookie_callback_py)
+
+        ftype_nv_cookie = ctypes.CFUNCTYPE(ctypes.c_void_p,    # Pointer to Python Function
+                                           ctypes.c_char_p)    # char *name
+
+        self.nv_cookie_callback_c = ftype_nv_cookie(self.nv_cookie_callback_py)
+
+        self.lib.handleExceptionParameters.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_longlong, ctypes.c_char_p,
+                                                       ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p,
+                                                       ctypes.c_char_p, ctypes.c_int, ctypes.c_char_p]
+
         self.logger.info("load common lib -> lib.obj :{0}".format(self.lib))
         
         return self.lib
 
+    def sdk_getNDSessionCookie(self, bt):
+        return self.lib.sdk_getNDSessionCookie(bt).decode()
+
+    def exception_dump_native(self, bt, mpp, starttime, excptnclsname, excptnmsg, throwingclsname,
+                        throwingmtdname, excptncause, excptnlineno, stacktrace):
+
+        #print("INSIDE AGENT PY ", excptnclsname)
+        self.lib.handleExceptionParameters(bt, mpp, ctypes.c_longlong(starttime), ctypes.c_char_p(excptnclsname.encode("utf-8")),
+                                           ctypes.c_char_p(excptnmsg.encode("utf-8")),
+                                           ctypes.c_char_p(throwingclsname.encode("utf-8")),
+                                           ctypes.c_char_p(throwingmtdname.encode("utf-8")),
+                                           ctypes.c_char_p(excptncause.encode("utf-8")),
+                                           excptnlineno,
+                                           ctypes.c_char_p(bytes(stacktrace, 'utf-8'))
+                                           )
+
+    def exceptiondump(self, bt, mpp, starttime, excptnclsname, excptnmsg, throwingclsname,
+                      throwingmtdname, excptncause, excptnlineno, stacktrace):
+        if self.cav_env == "NATIVE":
+            self.exception_dump_native(bt, mpp, starttime, excptnclsname, excptnmsg, throwingclsname,
+                                  throwingmtdname, excptncause, excptnlineno, stacktrace)
+        elif self.cav_env == "AWS_LAMBDA":
+            udp.exception_dump(self, bt, mpp, starttime, excptnclsname, excptnmsg, throwingclsname,
+                               throwingmtdname, excptncause, excptnlineno, stacktrace)
+
+    def nd_cookie_callback_py(self, name, domain, rheader, mpos, duration):
+        #print(f"nd_cookie name {name} \ndomain {domain}\nrheader {rheader}\nmpos {mpos}\nduration {duration}")
+        self.nd_cookie_key = name.decode()
+
+    def nv_cookie_callback_py(self, name):
+        #print(f"nv_cookie name {name}")
+        self.nv_cookie_key = name.decode()
 
     def get_trace_headers(self, bt):
         traceparent = self.lib.sdk_get_otl_traceparent(bt)
         tracestate = self.lib.sdk_get_otl_tracestate(bt)
         return traceparent, tracestate
 
-
     def http_req_resp_wrapper_native(self, bt, rbuffer, rtype, statuscode):
         rbuffer = str(rbuffer)
         rbuffer = bytes(rbuffer, 'utf-8')
@@ -212,7 +273,6 @@ class Agent(object):
 
         self.lib.sdk_httpReqRespWrapper(bt, r_buffer_c, r_type_c, statuscode)
 
-
     def http_req_resp_wrapper(self, bt, rbuffer, rtype, statuscode):
         if self.cav_env == "NATIVE":
             self.http_req_resp_wrapper_native(bt, rbuffer, rtype, statuscode)
@@ -221,7 +281,7 @@ class Agent(object):
 
 
     def ai_wrapper_begin_native(self):
-        print("************************ callback registered ********************************************")
+        #print("************************ callback registered ********************************************")
         self.lib.register_AI_start_callback(self.c_startaisession)
 
     def ai_wrapper_begin(self):
@@ -249,10 +309,24 @@ class Agent(object):
         #pathd = "/opt/cavisson/netdiagnostics/logs/"
         nd_home = os.environ.get('ND_HOME')
         pathd = nd_home + '/python/logs/'
-        filenameai1 = str(self.sess.decode('ascii')) + ".txt"
+        filenameai1 = str(self.sname.decode('ascii')) + ".txt"
         aifile_to_open_to_d = os.path.join(pathd, filenameai1)
-        self.lib.handleUploadDownload(aifile_to_open_to_d)
+        self.lib.handleUploadDownload(bytes(aifile_to_open_to_d, 'utf-8'))
 
+    def sdk_getTopoFpStr(self, bt):
+        topo = None
+        if bt:
+            topo = self.lib.sdk_getTopoFpStr(bt)
+            topo = topo.decode()
+        return topo
+
+    def register_nd_cookie_callback(self):
+        self.lib.register_nd_cookie_callback(self.nd_cookie_callback_c)
+        #print("**** callback registered for ND cookie ****")
+
+    def register_nv_cookie_callback(self):
+        self.lib.register_nv_cookie_callback(self.nv_cookie_callback_c)
+        #print("**** callback registered for NV cookie ****")
 
     def ad_wrapper(self):
         self.lib.register_AD_callback(self.c_startad)
@@ -294,8 +368,7 @@ class Agent(object):
 
     # @profile
     def method_entry_native(self, bt, method):
-        print('Inside method entry ', method, bt)
-        self.logger.info("method entry -> lib.obj :{0}".format(self.lib))
+        #print('Inside method entry ', method, bt)
         self.logger.info("method entry -> method name :{0}".format(method))
         # defer C.free(unsafe.Pointer(method_c))
         # method_bytes = bytes(method, 'utf-8')
@@ -321,16 +394,15 @@ class Agent(object):
 
     #@profile
     def method_exit_native(self, bt, method):
-            print('in method exit', method, bt)
-            self.logger.info("load common lib -> lib.obj :{0}".format(self.lib))
-            self.logger.info("method entry -> method name :{0}".format(method))
+            #print('in method exit', method, bt)
+            self.logger.info("method exit -> method name :{0}".format(method))
             # defer C.free(unsafe.Pointer(method_c))
             # method_bytes = bytes(method, 'utf-8')
             try:
                 method_bytes = bytes(method, 'utf-8')
             except:
                 method_bytes = bytes(method.encode("utf-8"))
-            print("\n\n", method_bytes)
+            #print("\n\n", method_bytes)
             self.lib.nd_method_exit(bt, ctypes.c_char_p(method_bytes))
 
     def method_exit(self, bt, method, status=200, duration=0):
@@ -357,34 +429,38 @@ class Agent(object):
 
 
     # @profile
-    def start_business_transaction_native(self, bt_name, correlation_header):
+    def start_business_transaction_native(self, bt_name, correlation_header, nd_cookie, nv_cookie, bt_header_value):
         if os.environ['nd_init_done'] == '0':
             self.start()
-            print("start_business_transaction ->")
-            self.logger.debug(" start_business_transaction -> lib.obj {0}:".format(self.lib))
+            #print("start_business_transaction ->")
+            #self.logger.debug(" start_business_transaction -> lib.obj {0}:".format(self.lib))
 
         try:
-            # self.logger.debug(" start_business_transaction -> lib.obj {0}:".format(self.lib))
+            self.logger.info(" start_business_transaction -> bt_name {0}:".format(bt_name))
             btname = bytes(bt_name, 'utf-8')
             # btname = bytes(bt_name.encode("utf-8"))
-            self.logger.info("btname type and value: {0} and {1}".format(type(btname), btname))
+            #self.logger.info("btname type and value: {0} and {1}".format(type(btname), btname))
             bt_name_c = ctypes.c_char_p(btname)
-            self.logger.info("bt_name_c value: {0}".format(bt_name_c))
+            #self.logger.info("bt_name_c value: {0}".format(bt_name_c))
             correlation_header = bytes(correlation_header, 'utf-8')
             # correlation_header = bytes(correlation_header.encode("utf-8"))
             correlation_header_c = ctypes.c_char_p(correlation_header)
         except:
             btname = bytes(bt_name.encode("utf-8"))
-            self.logger.info("btname type and value: {0} and {1}".format(type(btname), btname))
+            #self.logger.info("btname type and value: {0} and {1}".format(type(btname), btname))
             bt_name_c = ctypes.c_char_p(btname)
-            self.logger.info("bt_name_c value: {0}".format(bt_name_c))
+            #self.logger.info("bt_name_c value: {0}".format(bt_name_c))
             correlation_header = bytes(correlation_header.encode("utf-8"))
             correlation_header_c = ctypes.c_char_p(correlation_header)
 
         self.get_transaction_context().flow_path_id = generate_flow_path_id() 
         
-        nd_cookie_c = ctypes.c_char_p(None)
-        nv_cookie_c = ctypes.c_char_p(None)
+        nd_cookie_bytes = get_utf8_bytes(str(nd_cookie))
+        nd_cookie_c = ctypes.c_char_p(nd_cookie_bytes)
+
+        nv_cookie_bytes = get_utf8_bytes(str(nv_cookie))
+        nv_cookie_c = ctypes.c_char_p(nv_cookie_bytes)
+
         bt_header_value_c = ctypes.c_char_p(None)
         
         flow_path_id = self.get_transaction_context().flow_path_id
@@ -393,17 +469,17 @@ class Agent(object):
 
         bt = self.lib.nd_bt_begin(bt_name_c, correlation_header_c, nd_cookie_c, nv_cookie_c, bt_header_value_c, flow_path_id_c)
         #self.logger.info("bt datatype : {0}".format(type(bt)))
-        print("btname value: {0}".format(btname))
+        #print("btname value: {0}".format(btname))
         # print('-----------------\nCurrent heap memory stack for BT: ',bt_name,' is : ',hpy().heap())
         self.active_bts.add(bt)
         self.set_current_bt(bt)
-        print("BT value of this txn: ", bt)
+        #print("BT value of this txn: ", bt)
         return bt
 
-    def start_business_transaction(self, bt_name, correlation_header):
+    def start_business_transaction(self, bt_name, correlation_header, nd_cookie=None, nv_cookie=None, bt_header_value=None):
         bt = None
         if self.cav_env == "NATIVE":
-            bt = self.start_business_transaction_native(bt_name, correlation_header)
+            bt = self.start_business_transaction_native(bt_name, correlation_header, nd_cookie, nv_cookie, bt_header_value)
         elif self.cav_env == "AWS_LAMBDA":
             bt = udp.start_business_transaction(self, bt_name, correlation_header)
 
@@ -418,11 +494,10 @@ class Agent(object):
 
     # @profile
     def end_business_transaction_native(self, bt):
-        print("end_business_transaction--> BT: ", bt)
-        self.logger.info('end_business_transaction--> lib variable in end txn: {0}'.format(self.lib))
+        #print("end_business_transaction--> BT: ", bt)
         self.logger.info('At end_business_transaction bt value is {0}'.format(bt))
         statuscode = self.get_current_status_code()
-        rc = self.lib.nd_bt_end(bt, statuscode)
+        rc = self.lib.nd_bt_end(bt, statuscode, ctypes.c_int(0))
         self.logger.info("nd_bt_end return value:  {0}".format(int(rc)))
         self.unset_current_bt()
         self.unset_current_status_code()
@@ -434,6 +509,7 @@ class Agent(object):
             rc = self.end_business_transaction_native(bt)
         elif self.cav_env == "AWS_LAMBDA":
             rc = udp.end_business_transaction(self, bt)
+        self.get_transaction_context().havoc_applied = False
         return rc
 
 
