@@ -8,6 +8,7 @@ import time
 import logging
 logger = logging.getLogger('pythonagent.agent')
 
+
 class Header(object):
 
     def __init__(self, context, apiReqId, awsReqId, funcName):
@@ -253,7 +254,7 @@ class MethodExitMessage(object):
         self.eventType = 0  # DUMMY
         self.isCallout = 1  # DUMMY 1
         self.threadId = 0  # DUMMY threading.get_ident()
-        self.flow_path_instance = flow_path_instance 
+        self.flow_path_instance = flow_path_instance
         self.cpuTime = 0  # DUMMY
 
         self.len_methodName = None
@@ -339,7 +340,7 @@ class MethodExitMessage(object):
 
 class EndTransactionMessage(object):
 
-    def __init__(self, status_code, flow_path_instance):
+    def __init__(self, status_code, flow_path_instance, trace_request):
 
         self.status_code = status_code
         self.flow_path_instance = flow_path_instance
@@ -348,6 +349,7 @@ class EndTransactionMessage(object):
         self.header_len = None
         self.total_len = None
         self.msg_type = 3  # END TRANSACTION
+        self.trace_request = trace_request
 
     def encode(self):
 
@@ -366,9 +368,10 @@ class EndTransactionMessage(object):
                           + 8                                      # flow_path_instance(long long)
                           + 8                                      # endTime(long long)
                           + 8                                      # cpuTime(long long)
+                          + 4                                      # trace_request(int)
                           + 1)                                     # \n (1c)
 
-        end_transaction_message_struct = struct.Struct("=ciiiciqqqc")
+        end_transaction_message_struct = struct.Struct("=ciiiciqqqic")
 
         end_transaction_message_values = (bytes('^', "utf-8"),
                                           self.header_len,
@@ -379,6 +382,7 @@ class EndTransactionMessage(object):
                                           self.flow_path_instance,
                                           self.endTime,
                                           self.cpuTime,
+                                          self.trace_request,
                                           bytes('\n', "utf-8"))
 
         end_transaction_message_packet = end_transaction_message_struct.pack(*end_transaction_message_values)
@@ -388,7 +392,7 @@ class EndTransactionMessage(object):
 
 class TransactionEncodeHttpMessage(object):
 
-    def __init__(self, bt, rbuffer, rtype, statuscode, flow_path_instance):
+    def __init__(self, bt, rbuffer, rtype, statuscode, flow_path_instance, otl_trace_id="", otl_parent_id="", otl_trace_state=""):
 
         # bt = 140544547110928
         # http_host = "www.google.com"
@@ -397,16 +401,52 @@ class TransactionEncodeHttpMessage(object):
         self.type = rtype
         self.statuscode = statuscode
         self.flow_path_instance = flow_path_instance
+        self.otl_trace_id = otl_trace_id
+        self.otl_parent_id = otl_parent_id
+        self.otl_trace_state = otl_trace_state
+        self.otl_version = 1
+        self.otl_trace_flag = 0
+
 
     def encode(self):
-        self.buffer_len = len(self.buffer)
-        self.type_len = len(self.type)
+        self.len_buffer = len(self.buffer)
+        self.len_type = len(self.type)
+        self.len_otl_trace_id = len(self.otl_trace_id)
+        self.len_otl_parent_id = len(self.otl_parent_id)
+        self.len_otl_trace_state = len(self.otl_trace_state)
 
-        self.header_len = 12
-        self.total_len = 12 + 20 + self.buffer_len + self.type_len + 3
-        self.msg_type = 6  # HTTP CALLOUT BEGIN, 7 FOR END
+        self.header_len = (4                                       # header_len(int)
+                           + 4                                     # total_len(int)
+                           + 4)                                    # msg_type(int)
 
-        format_str = "=ciiiciiiq{}s{}sc".format(self.buffer_len, self.type_len)
+        self.total_len = (1                                        # ^ (1c)
+                          + 4                                      # header_len(int)
+                          + 4                                      # total_len(int)
+                          + 4                                      # msg_type(int)
+                          + 1                                      # | (1c)
+                          + 4                                      # status_code(int)
+                          + 4                                      # buffer_len(int)
+                          + 4                                      # type_len(int)
+                          + 4                                      # otl_version(int)
+                          + 4                                      # otl_trace_flag(int)
+                          + 4                                      # len_otl_trace_id(int)
+                          + 4                                      # len_otl_parent_id(int)
+                          + 4                                      # len_otl_trace_state(int)
+                          + 8                                      # flow_path_instance(long long)
+                          + self.len_buffer                        # variable
+                          + self.len_type                          # variable
+                          + self.len_otl_trace_id                  # variable
+                          + self.len_otl_parent_id                 # variable
+                          + self.len_otl_trace_state               # variable
+                          + 1)                                     # \n (1c)
+
+        self.msg_type = 6
+
+        format_str = "=ciiiciiiiiiiiq{}s{}s{}s{}s{}sc".format(self.len_buffer,
+                                                              self.len_type,
+                                                              self.len_otl_trace_id,
+                                                              self.len_otl_parent_id,
+                                                              self.len_otl_trace_state)
 
         transaction_encode_http_message_struct = struct.Struct(format_str)
 
@@ -416,11 +456,19 @@ class TransactionEncodeHttpMessage(object):
                                                   self.msg_type,
                                                   bytes('|', "utf-8"),
                                                   self.statuscode,
-                                                  self.buffer_len,
-                                                  self.type_len,
+                                                  self.len_buffer,
+                                                  self.len_type,
+                                                  self.otl_version,
+                                                  self.otl_trace_flag,
+                                                  self.len_otl_trace_id,
+                                                  self.len_otl_parent_id,
+                                                  self.len_otl_trace_state,
                                                   self.flow_path_instance,
                                                   self.buffer.encode('utf-8'),
                                                   self.type.encode('utf-8'),
+                                                  self.otl_trace_id.encode('utf-8'),
+                                                  self.otl_parent_id.encode('utf-8'),
+                                                  self.otl_trace_state.encode('utf-8'),
                                                   bytes('\n', "utf-8"))
 
         transaction_encode_http_message_packet = transaction_encode_http_message_struct.pack(*transaction_encode_http_message_values)
@@ -428,6 +476,102 @@ class TransactionEncodeHttpMessage(object):
 
         return transaction_encode_http_message_packet
 
+class ExceptionEncodeMessage(object):
+
+    def __init__(self,
+                 line_number,
+                 exception_class_name,
+                 exception_throwing_class_name,
+                 exception_throwing_method_name,
+                 exception_message,
+                 exception_cause,
+                 exception_stack_trace,
+                 flow_path_instance,
+                 start_time):
+
+        self.line_number = line_number
+        self.exception_class_name = exception_class_name 
+        self.exception_throwing_class_name = exception_throwing_class_name 
+        self.exception_throwing_method_name = exception_throwing_method_name
+        self.exception_message = exception_message
+        self.exception_cause = exception_cause
+        self.exception_stack_trace = exception_stack_trace
+        self.flow_path_instance = flow_path_instance 
+        self.start_time = start_time
+
+    def encode(self):
+        self.len_of_exception_class_name = len(self.exception_class_name)
+        self.len_of_exception_throwing_class_name = len(self.exception_throwing_class_name)
+        self.len_of_exception_throwing_method_name = len(self.exception_throwing_method_name)
+        self.len_of_exception_message = len(self.exception_message)
+        self.len_of_exception_cause = len(self.exception_cause)
+        self.len_of_exception_stack_trace = len(self.exception_stack_trace)
+
+        self.header_len = (4                                             # header_len(int)
+                           + 4                                           # total_len(int)
+                           + 4)                                          # msg_type(int)
+
+        self.total_len = (1                                              # ^ (1c)
+                          + 4                                            # header_len(int)
+                          + 4                                            # total_len(int)
+                          + 4                                            # msg_type(int)
+                          + 1                                            # | (1c)
+                          + 4                                            # int lineNumber 
+                          + 4                                            # int lenofExceptionClassName 
+                          + 4                                            # int lenofExceptionThrowingClassName 
+                          + 4                                            # int lenofExceptionThrowingMethodName
+                          + 4                                            # int lenofExceptionMessage
+                          + 4                                            # int lenofExceptionCause
+                          + 4                                            # int lenofExceptionStackTrace
+                          + 8                                            # long long flowpathInstance; 
+                          + 8                                            # long long startTime 
+                          + self.len_of_exception_class_name             # char* ExceptionClassName 
+                          + self.len_of_exception_throwing_class_name    # char* ExceptionThrowingClassName; 
+                          + self.len_of_exception_throwing_method_name   # char* ExceptionThrowingMethodName; 
+                          + self.len_of_exception_message                # char* ExceptionMessage; 
+                          + self.len_of_exception_cause                  # char* ExceptionCause; 
+                          + self.len_of_exception_stack_trace            # char* ExceptionStackTrace;
+                          + 1)                                           # \n (1c)
+
+        self.msg_type = 5
+
+
+        format_str = "=ciiiciiiiiiiqq{}s{}s{}s{}s{}s{}sc".format(self.len_of_exception_class_name,
+                                                                 self.len_of_exception_throwing_class_name,
+                                                                 self.len_of_exception_throwing_method_name,
+                                                                 self.len_of_exception_message,
+                                                                 self.len_of_exception_cause,
+                                                                 self.len_of_exception_stack_trace)
+
+        exception_encode_message_struct = struct.Struct(format_str)
+
+
+        exception_encode_message_values = (bytes('^', "utf-8"),
+                                           self.header_len,
+                                           self.total_len,
+                                           self.msg_type,
+                                           bytes('|', "utf-8"),
+                                           self.line_number,
+                                           self.len_of_exception_class_name,
+                                           self.len_of_exception_throwing_class_name,
+                                           self.len_of_exception_throwing_method_name,
+                                           self.len_of_exception_message,
+                                           self.len_of_exception_cause,
+                                           self.len_of_exception_stack_trace,
+                                           self.flow_path_instance,
+                                           self.start_time,
+                                           self.exception_class_name.encode('utf-8'),
+                                           self.exception_throwing_class_name.encode('utf-8'),
+                                           self.exception_throwing_method_name.encode('utf-8'),
+                                           self.exception_message.encode('utf-8'),
+                                           self.exception_cause.encode('utf-8'),
+                                           self.exception_stack_trace.encode('utf-8'),
+                                           bytes('\n', "utf-8"))
+
+        exception_encode_message_packet = exception_encode_message_struct.pack(*exception_encode_message_values)
+
+        return exception_encode_message_packet
+
 
 class HavocMessage(object):
 
@@ -491,8 +635,12 @@ def create_method_exit_message(context, bt, method, backend_header, status, dura
 def create_end_transaction_message(context, bt, statuscode):
     header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
     header_packet = header.encode()
+    if context.havoc_applied:
+        trace_request = 4
+    else:
+        trace_request = 0
 
-    end_transaction_message = EndTransactionMessage(statuscode, context.flow_path_id)
+    end_transaction_message = EndTransactionMessage(statuscode, context.flow_path_id, trace_request)
     end_transaction_message_packet = end_transaction_message.encode()
 
     full_packet = header_packet + end_transaction_message_packet
@@ -525,6 +673,29 @@ def create_transaction_encode_http_message(context, bt, rbuffer, rtype, statusco
 
     return full_packet
 
+def create_exception_encode_message(context, line_number, exception_class_name, 
+                                    exception_throwing_class_name, exception_throwing_method_name, 
+                                    exception_message, exception_cause, exception_stack_trace, start_time):
+
+    header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
+    header_packet = header.encode()
+
+
+    exception_encode_message = ExceptionEncodeMessage(line_number, 
+                                                      exception_class_name,
+                                                      exception_throwing_class_name,
+                                                      exception_throwing_method_name,
+                                                      exception_message,
+                                                      exception_cause,
+                                                      exception_stack_trace,
+                                                      context.flow_path_id,
+                                                      start_time)
+
+    exception_encode_message_packet = exception_encode_message.encode()
+
+    full_packet = header_packet + exception_encode_message_packet
+
+    return full_packet
 
 class InitMessage(object):
 
@@ -537,6 +708,7 @@ class InitMessage(object):
         init_message_packet = init_message_struct.pack(*init_message_values)
         return init_message_packet
 
+
 def create_init_message(context):
     header = Header(context, context.api_request_id, context.aws_request_id, context.function_name)
     logger.debug('function name inside wrap header {}'.format(context.function_name))
