@@ -6,6 +6,7 @@ from pythonagent.agent.internal import udp
 from pythonagent.agent.probes.havoc.havoc_constants import HAVOC_TYPES, EXPIRE_DURATION_MS, SIZE_OF_OBJECTS_BYTES, HAVOC_PROFILES_JSON
 from pythonagent.agent.probes.havoc.custom_memory_stress import apply_memory_leak
 from pythonagent.utils import get_current_timestamp_in_ms
+import pythonagent.agent as agent
 
 havoc_types = {
     1: "InBoundService_Delay",
@@ -304,13 +305,15 @@ class NDNetHavocMonitor(object):
         self.logger.info("Applying delay duration(sec): {}".format(total_duration_in_sec))
         try:
             time.sleep(total_duration_in_sec)
+            _agent = agent.get_agent_instance()
+            _agent.get_transaction_context().havoc_applied = True
         except Exception as e:
             raise e
 
     def common_failure(self, exception_message):
         self.logger.info("Applying failure message: {}".format(exception_message))
-
-
+        _agent = agent.get_agent_instance()
+        _agent.get_transaction_context().havoc_applied = True
         raise NDHavocException(exception_message)
 
     @staticmethod
@@ -455,6 +458,18 @@ class NDNetHavocMonitor(object):
 
                 if method_fqm in req_method_fqm:
                     self.logger.debug("Config matched for Request FQM {}".format(req_method_fqm))
+                    _agent = agent.get_agent_instance()
+                    start_time = round(time.time() * 1000)
+                    throwing_method = req_method_fqm
+                    dummy_bt = 0
+                    dummy_line_number = 0
+                    encodedstktrc = "DummyStackTrace"
+                    exception_class = "HavocException"
+                    exception_message = "HavocMessage"
+                    exception_cause = "HavocMethodFailure"
+                    throwing_class = "None"
+                    _agent.exceptiondump(dummy_bt, None, start_time, exception_class, exception_message, throwing_class,
+                                         throwing_method, exception_cause, dummy_line_number, encodedstktrc)
                     self.common_failure(method_fqm)
                 else:
                     self.logger.debug("Config not matched for Request FQM {}".format(req_method_fqm))
@@ -471,6 +486,8 @@ class NDNetHavocMonitor(object):
                 totalDurationInSec = config.totalDurationInSec
                 memoryLeaksInSec = config.memoryLeaksInSec
                 apply_memory_leak(leaksINMB, totalDurationInSec, memoryLeaksInSec)
+                _agent = agent.get_agent_instance()
+                _agent.get_transaction_context().havoc_applied = True
 
         else:
             self.logger.debug("config not set for apply_method_call_failure")
