@@ -27,7 +27,7 @@ class HttplibConnectionInterceptor(HTTPConnectionInterceptor):
             #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest connection is :{0}".format(connection))
             #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest putrequest is :{0}".format(putrequest))
             host_port = str(connection.host) + "$" + str(connection.port)
-            exit_call = self.http_call_begin(bt, host_port, url)
+            exit_call = self.http_call_begin(bt, host_port, url, "http.client.HTTPConnection.getresponse")
             #self.agent.logger.info("Modulename: HttplibConnectionInterceptor class inside _putrequest exit_call is :{0}".format(exit_call))
 
             if bt:
@@ -70,9 +70,7 @@ class HttplibConnectionInterceptor(HTTPConnectionInterceptor):
             with self.end_exit_call_and_reraise_on_exception(exit_call,
                                                              ignored_exceptions=(TypeError,)):
                 #self.agent.logger.debug("inside with")
-                self.agent.method_entry(bt, "getresponse")
                 response = getresponse(connection, *args, **kwargs)
-                self.agent.method_exit(bt, "getresponse")
                 #print("RESPONSE STATUS CODE: ", response.status)
         except TypeError:
             #print("inside except")
@@ -81,7 +79,7 @@ class HttplibConnectionInterceptor(HTTPConnectionInterceptor):
                 response = getresponse(connection)
                 #print("RESPONSE STATUS CODE: ", response.status)
 
-        self.http_call_end(self.bt, exit_call, response.status)
+        self.http_call_end(self.bt, exit_call, "http.client.HTTPConnection.getresponse", response.status)
         #self.proxy.http_call_end(self.agent.bt,exit_call)
         try:
             del connection._pythonagent_exit_call
