@@ -2,11 +2,6 @@ import sys
 import os
 import ast
 import json
-import logging
-
-
-logger = logging.getLogger('pythonagent.agent')
-
 
 #path ="C:/Users/garima.singh/PycharmProjects/django_redis_demo"
 path ="/home/cavisson/shop/my-shop/pythonagent_garima"
@@ -20,8 +15,8 @@ for root, dirs, files  in os.walk(path):
             modules.append(file)
             default_path.append(os.path.join(root,file))
            
-logger.info("modules {}".format(modules))
-logger.info("default_path{}".format(default_path))
+print("modules",modules)
+print("default_path",default_path)
 
 
 data = {
