@@ -2,7 +2,7 @@
 import errno
 import os
 #from agent.internal.proxy import Proxy
-from main.pytrace import CommandInvocationError, CommandExecutionError
+from pythonagent.main.pytrace import CommandInvocationError, CommandExecutionError
 import sys
 
 USAGE = "run <command> [args...]"
